// ----------------------------------------------------------------------------
// Copyright (C) 2003-2005 Microsoft Corporation, All rights reserved.
// ----------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Threading;

namespace Microsoft.ServiceModel.Samples
{
    /// <summary>
    /// Handles asynchronous interactions between producers and consumers. 
    /// Producers can dispatch available data to the input queue, 
    /// where it will be dispatched to a waiting consumer or stored until a
    /// consumer becomes available. Consumers can synchronously or asynchronously
    /// request data from the queue, which will be returned when data becomes
    /// available.
    /// </summary>
    /// <typeparam name="T">The concrete type of the consumer objects that are waiting for data.</typeparam>
    class InputQueue<T> : IDisposable where T : class, IDisposable
    {
        //Stores items that are waiting to be consumed.
        ItemQueue itemQueue;

        //Each IQueueReader represents some consumer that is waiting for
        //items to appear in the queue. The readerQueue stores them
        //in an ordered list so consumers get serviced in a FIFO manner.
        Queue<IQueueReader> readerQueue;

        //Represents the current state of the InputQueue
        //as it transitions through its lifecycle.
        QueueState queueState;
        enum QueueState
        {
            Created,
            Open,
            Closed
        }

        public InputQueue()
        {
            itemQueue = new ItemQueue();
            readerQueue = new Queue<IQueueReader>();
            queueState = QueueState.Created;
        }

        public int PendingCount
        {
            get
            {
                lock (ThisLock)
                {
                    return itemQueue.ItemCount;
                }
            }
        }

        object ThisLock
        {
            get { return itemQueue; }
        }

        //Opens the InputQueue and allows it to start
        //dispatching items to consumers
        public void Open()
        {
            lock (ThisLock)
            {
                if (queueState == QueueState.Open)
                {
                    return;
                }

                if (queueState == QueueState.Closed)
                    throw new ObjectDisposedException(this.ToString());

                if (queueState == QueueState.Created)
                {
                    queueState = QueueState.Open;
                }
            }
        }

        //Begins an asynchronous Dequeue() operation. If an item is available we'll
        //return it immediately. Otherwise, we'll associate an AsyncQueueReader
        //with the callback and store it in the readerQueue.
        public IAsyncResult BeginDequeue(AsyncCallback callback, object state)
        {
            Item item;

            lock (ThisLock)
            {
                if (queueState == QueueState.Open)
                {
                    //Callback with the first available item if one
                    //is available.
                    if (itemQueue.HasAvailableItem)
                    {
                        item = itemQueue.DequeueAvailableItem();
                        return new TypedCompletedAsyncResult<T>(item.Value, callback, state);
                    }
                    //Otherwise, represent the request with an AsyncQueueReader
                    //which will call back with the item when it becomes available.
                    else
                    {
                        AsyncQueueReader reader = new AsyncQueueReader(callback, state);
                        readerQueue.Enqueue(reader);
                        return reader;
                    }
                }
                else if (queueState == QueueState.Closed)
                {
                    throw new ObjectDisposedException(this.ToString());
                }
                else // queueState == QueueState.Created 
                {
                    throw new InvalidOperationException();
                }
            }
        }

        //Ends an asynchronous Dequeue operation.
        public T EndDequeue(IAsyncResult result)
        {
            TypedCompletedAsyncResult<T> typedResult = result as TypedCompletedAsyncResult<T>;
            if (typedResult != null)
                return TypedCompletedAsyncResult<T>.End(result);

            return AsyncQueueReader.End(result);
        }

        //A synchronous Dequeue operation that uses WaitQueueReaders instead
        //of AsyncQueueReaders.
        public T Dequeue()
        {
            WaitQueueReader reader;
            Item item;

            lock (ThisLock)
            {
                if (queueState == QueueState.Open)
                {
                    //If an item is available, return it immediately.
                    if (itemQueue.HasAvailableItem)
                    {
                        item = itemQueue.DequeueAvailableItem();
                        reader = null;
                    }
                    //Otherwise, represent the request with
                    //a WaitQueueReader and store it for later use.
                    else
                    {
                        item = new Item();
                        reader = new WaitQueueReader();
                        readerQueue.Enqueue(reader);
                    }
                }
                else if (queueState == QueueState.Closed)
                {
                    throw new ObjectDisposedException(this.ToString());
                }
                else // queueState == QueueState.Created
                {
                    throw new InvalidOperationException();
                }
            }

            //If we don't have an item, block on the WaitQueueReader
            //until data becomes available to return.
            if (reader != null)
            {
                return reader.Wait();
            }
            else
            {
                //Return the value of the item that was
                //immediately available
                return item.Value;
            }
        }

        //Enqueues some data into the InputQueue and hands it off
        //to a waiting reader (if we have one). If nobody's waiting for this
        //data yet, hold on to it in the itemQueue until someone asks for it.
        public void EnqueueAndDispatch(T item)
        {
            if (item == null)
                throw new ArgumentNullException("item");

            bool disposeItem = false;
            IQueueReader reader = null;

            lock (ThisLock)
            {
                if (queueState == QueueState.Open)
                {
                    if (readerQueue.Count > 0)
                    {
                        reader = readerQueue.Dequeue();
                    }
                    else
                    {
                        itemQueue.EnqueueAvailableItem(item);
                    }
                }
                else if (queueState == QueueState.Closed)
                {
                    disposeItem = true;
                }
                else // queueState == QueueState.Created
                {
                    itemQueue.EnqueueAvailableItem(item);
                }
            }

            if (reader != null)
            {
                reader.Set(item);
            }
            else if (disposeItem)
            {
                item.Dispose();
            }
        }

        #region IDisposable

        public void Close()
        {
            ((IDisposable)this).Dispose();
        }

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        protected void Dispose(bool disposing)
        {
            if (disposing)
            {
                bool dispose = false;

                lock (ThisLock)
                {
                    if (queueState != QueueState.Closed)
                    {
                        queueState = QueueState.Closed;
                        dispose = true;
                    }
                }

                if (dispose)
                {
                    while (readerQueue.Count > 0)
                    {
                        IQueueReader reader = readerQueue.Dequeue();
                        reader.Set(null);
                    }

                    while (itemQueue.HasAnyItem)
                    {
                        Item item = itemQueue.DequeueAnyItem();
                        item.Value.Dispose();
                    }
                }
            }
        }
        #endregion

        #region Queue Reader Implementations
        interface IQueueReader
        {
            void Set(T item);
        }

        class WaitQueueReader : IQueueReader
        {
            T item;
            ManualResetEvent waitEvent;

            public WaitQueueReader()
            {
                waitEvent = new ManualResetEvent(false);
            }

            public void Set(T item)
            {
                lock (this)
                {
                    Debug.Assert(this.item == null);
                    this.item = item;
                    waitEvent.Set();
                }
            }

            public T Wait()
            {
                waitEvent.WaitOne();
                return item;
            }
        }

        class AsyncQueueReader : AsyncResult, IQueueReader
        {
            T item;

            public AsyncQueueReader(AsyncCallback callback, object state)
                : base(callback, state)
            {
            }

            public static T End(IAsyncResult result)
            {
                AsyncQueueReader readerResult = result as AsyncQueueReader;
                if (readerResult == null)
                    throw new ArgumentException("result");
                AsyncResult.End(readerResult);

                return readerResult.item;
            }

            public void Set(T item)
            {
                lock (this)
                {
                    this.item = item;
                }

                Complete(false);
            }
        }
        #endregion

        #region Items and ItemQueue
        struct Item
        {
            T value;

            public Item(T value)
            {
                this.value = value;
            }

            public T Value
            {
                get { return value; }
            }
        }

        class ItemQueue
        {
            Item[] items;
            int head;
            int pendingCount;
            int totalCount;

            public ItemQueue()
            {
                items = new Item[1];
            }

            public bool HasAvailableItem
            {
                get { return totalCount > pendingCount; }
            }

            public bool HasAnyItem
            {
                get { return totalCount > 0; }
            }

            public int ItemCount
            {
                get { return totalCount; }
            }

            public int PendingItemCount
            {
                get { return pendingCount; }
            }

            public Item DequeueAvailableItem()
            {
                if (totalCount == pendingCount)
                    Debug.Assert(false, "ItemQueue does not contain any available items");
                return DequeueItemCore();
            }

            public Item DequeueAnyItem()
            {
                if (pendingCount == totalCount)
                    pendingCount--;
                return DequeueItemCore();
            }

            Item DequeueItemCore()
            {
                if (totalCount == 0)
                    Debug.Assert(false, "ItemQueue does not contain any items");
                Item item = items[head];
                items[head] = new Item();
                totalCount--;
                head = (head + 1) % items.Length;
                return item;
            }

            public void EnqueueAvailableItem(T value)
            {
                if (totalCount == items.Length)
                {
                    Item[] newItems = new Item[items.Length * 2];
                    for (int i = 0; i < totalCount; i++)
                        newItems[i] = items[(head + i) % items.Length];
                    head = 0;
                    items = newItems;
                }
                int tail = (head + totalCount) % items.Length;
                items[tail] = new Item(value);
                totalCount++;
            }
        }
        #endregion
    }
}
