// ----------------------------------------------------------------------------
// Copyright (C) 2003-2005 Microsoft Corporation, All rights reserved.
// ----------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.ServiceModel;
using System.ServiceModel.Channels;
using System.Globalization;

namespace Microsoft.ServiceModel.Samples
{
    /// <summary>
    /// IChannelFactory implementation for Udp.
    /// 
    /// Supports IOutputChannel only, as Udp is fundamentally
    /// a datagram protocol.
    /// </summary>
    public class UdpChannelFactory : ChannelFactoryBase
    {
        IBufferManager bufferManager;

        public UdpChannelFactory()
            : base()
        {
        }

        internal IBufferManager BufferManager
        {
            get
            {
                return this.bufferManager;
            }
        }

        /// <summary>
        /// We only support the default Message Version
        /// </summary>
        public override MessageVersion MessageVersion
        {
            get
            {
                return UdpConstants.MessageVersion;
            }
        }

        /// <summary>
        /// URI scheme associated with the UDP transport 
        /// </summary>
        public override string Scheme
        {
            get
            {
                return UdpConstants.Scheme;
            }
        }

        /// <summary>
        /// Used by higher layers to determine what types of channels this
        /// factory supports. Which in this case is just IOutputChannel.
        /// </summary>
        public override bool CanCreateChannel<TChannel>()
        {
            return (typeof(TChannel) == typeof(IOutputChannel));
        }

        /// <summary>
        /// Create a new Udp Channel. Supports IOutputChannel.
        /// </summary>
        /// <typeparam name="TChannel">The type of Channel to create (e.g. IOutputChannel)</typeparam>
        /// <param name="remoteAddress">The address of the remote endpoint</param>
        /// <returns></returns>
        protected override TChannel OnCreateChannel<TChannel>(EndpointAddress remoteAddress)
        {
            if (typeof(TChannel) == typeof(IOutputChannel))
            {
                return (TChannel)(object)new UdpOutputChannel(this, remoteAddress);
            }

            throw new ArgumentException(string.Format(CultureInfo.CurrentCulture, 
                "Unsupported channel type: {0}.", typeof(TChannel).Name), "TChannel");
        }

        protected override void OnOpen()
        {
            // create a buffer manager to re-use byte arrays as necessary for performance
            this.bufferManager = new BufferManager(UdpConstants.MaxBufferSize * 10, UdpConstants.MaxBufferSize);
        }

        protected override void OnClosed()
        {
            if (this.bufferManager != null)
            {
                this.bufferManager.Clear();
                this.bufferManager = null;
            }
        }
    }
}
