// ----------------------------------------------------------------------------
// Copyright (C) 2003-2005 Microsoft Corporation, All rights reserved.
// ----------------------------------------------------------------------------

using System;
using System.Diagnostics;
using System.Collections;
using System.Collections.Generic;

using System.ServiceModel;
using System.ServiceModel.Channels;

namespace Microsoft.ServiceModel.Samples
{

    //The UdpInputListener is a UDP-specific implementation of the
    //protocol-agnostic IListener interface. Listeners are the Indigo 
    //analog to BSD sockets. 
    class UdpInputListener : ListenerBase<IInputChannel>
    {
        //Although we maintain a InputQueue of channels associated with
        //this listener, this particular implementation allows exactly one
        //channel instance per listener instance.
        InputQueue<IInputChannel> channelQueue;

        //The channel associated with this listener.
        UdpInputChannel currentChannel;

        //The listener factory that created this listener.
        UdpListenerFactory factory;

        object currentChannelLock;

        internal UdpInputListener(UdpListenerFactory factory, EndpointAddress localAddress, Filter filter,
            int priority)
            : base(factory, localAddress, filter, priority)
        {
            this.channelQueue = new InputQueue<IInputChannel>();
            this.channelQueue.Open();
            this.factory = factory;
            this.currentChannelLock = new object();
        }

        public override int PendingChannelCount
        {
            get
            {
                return this.channelQueue.PendingCount;
            }
        }

        //Synchronously returns a channel that is attached to this listener.
        public override IInputChannel AcceptChannel()
        {
            if (!this.IsDisposed)
            {
                this.EnsureChannelAvailable();
            }

            return this.channelQueue.Dequeue();
        }

        public override IAsyncResult BeginAcceptChannel(AsyncCallback callback, object state)
        {
            return this.channelQueue.BeginDequeue(callback, state);
        }

        public override IInputChannel EndAcceptChannel(IAsyncResult result)
        {
            return this.channelQueue.EndDequeue(result);
        }

        //Guarantees that channel is attached to this listener.
        private void EnsureChannelAvailable()
        {
            UdpInputChannel newChannel;
            bool channelCreated = CreateOrRetrieveChannel(out newChannel);

            if (channelCreated)
            {
                this.channelQueue.EnqueueAndDispatch(newChannel);
            }
        }

        //Dispatches a message to a waiting channel, creating one if necessary.
        internal void Dispatch(Message message)
        {
            UdpInputChannel newChannel;
            bool channelCreated = CreateOrRetrieveChannel(out newChannel);

            newChannel.Dispatch(message);

            if (channelCreated)
            {
                //Hand the channel off to whomever is waiting for AcceptChannel()
                //to complete
                this.channelQueue.EnqueueAndDispatch(newChannel);
            }
        }

        private bool CreateOrRetrieveChannel(out UdpInputChannel newChannel)
        {
            bool channelCreated = false;

            if ((newChannel = currentChannel) == null)
            {
                lock (currentChannelLock)
                {
                    if ((newChannel = currentChannel) == null)
                    {
                        newChannel = new UdpInputChannel((UdpListenerFactory)this.Factory, this.LocalAddress);
                        newChannel.Closed += new EventHandler(this.OnChannelClosed);
                        currentChannel = newChannel;
                        channelCreated = true;
                    }
                }
            }

            return channelCreated;
        }

        private void OnChannelClosed(object sender, EventArgs args)
        {
            UdpInputChannel channel = (UdpInputChannel)sender;

            lock (this.currentChannelLock)
            {
                if (channel == this.currentChannel)
                {
                    this.currentChannel = null;
                }
            }
        }

        protected override void OnDispose()
        {
            try
            {
                base.OnDispose();
            }
            finally
            {
                try
                {
                    factory.UnregisterListener(this);
                }
                finally
                {
                    channelQueue.Dispose();
                }
            }
        }
    }
}