// ----------------------------------------------------------------------------
// Copyright (C) 2003-2005 Microsoft Corporation, All rights reserved.
// ----------------------------------------------------------------------------

using System;
using System.IO;
using System.Xml;
using System.Diagnostics;
using System.Net;
using System.Net.Sockets;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Globalization;
using System.ServiceModel;
using System.ServiceModel.Channels;
using System.Threading;

namespace Microsoft.ServiceModel.Samples
{
    /// <summary>
    /// IListenerFactory implementation for Udp.
    /// 
    /// Supports IInputChannel only, as Udp is fundamentally
    /// a datagram protocol. Uses a listening UDP socket for each
    /// Factory instance, then demuxes to the appropriate IInputListener
    /// based on the registered filters. More advanced implementations could factor
    /// out the listening socket and demux from this listening object
    /// to separate factories based on a framing Via.
    /// </summary>
    public class UdpListenerFactory : ListenerFactoryBase
    {
        IBufferManager bufferManager;

        //The UDP network socket.
        Socket listenSocket;

        // Many Listeners can be registered for incoming data over
        // over this factory's socket. This filter table allows us 
        // to demultiplex Messages from our socket to the appropriate Udp Listener
        FilterTable<UdpInputListener> listenerTable;

        AsyncCallback onReceive;
        Uri uri;

        public UdpListenerFactory()
            : base()
        {
        }

        public override MessageVersion MessageVersion
        {
            get
            {
                return UdpConstants.MessageVersion;
            }
        }

        public override string Scheme
        {
            get
            {
                return UdpConstants.Scheme;
            }
        }

        public override Uri Uri
        {
            get
            {
                return this.uri;
            }
        }

        /// <summary>
        /// Used by higher layers to determine what types of channels this
        /// factory supports. Which in this case is just IInputChannel.
        /// </summary>
        public override bool CanCreateListener<TChannel>()
        {
            return (typeof(TChannel) == typeof(IInputChannel));
        }

        /// <summary>
        /// Create a new listener for this factory
        /// </summary>
        /// <typeparam name="TChannel">The type of listener to create</typeparam>
        /// <param name="localAddress">Local address for this listener (used in conjunction with filter to match messages)</param>
        /// <param name="filter">Filter to match incoming messages</param>
        /// <param name="priority">Priority of the filter (higher number == higher priority)</param>
        /// <returns>A new IListener</returns>
        protected override IListener<TChannel> OnCreateListener<TChannel>(EndpointAddress localAddress, 
            Filter filter, int priority)
        {
            if (typeof(TChannel) == typeof(IInputChannel))
            {
                if (localAddress != null)
                {
                    filter = new EndpointFilter(localAddress, filter);
                }

                UdpInputListener listener = new UdpInputListener(this, localAddress, filter, priority);

                lock (this.ThisLock)
                {
                    this.listenerTable.Add(filter, listener, priority);
                }

                return (IListener<TChannel>)(object)listener;
            }

            throw new ArgumentException(string.Format(CultureInfo.CurrentCulture, "Unsupported channel type: {0}.", typeof(TChannel).Name));
        }

        #region Lifecycle State Machine
        /// <summary>
        /// Shutdown ungracefully
        /// </summary>
        protected override void OnAbort()
        {
            try
            {
                // Abort can be called at anytime, so we can't assume that
                // we've been Opened successfully (and thus may not have a listen socket)
                if (listenSocket != null)
                {
                    this.listenSocket.Close(0);
                }
            }
            finally
            {
                base.OnAbort();
            }
        }

        /// <summary>
        /// Shutdown gracefully
        /// </summary>
        protected override void OnClose()
        {
            this.listenSocket.Close();
            base.OnClose();
        }

        protected override void OnClosed()
        {
            if (this.bufferManager != null)
            {
                this.bufferManager.Clear();
                this.bufferManager = null;
            }
        }

        /// <summary>
        /// Initialize any objects we're going to need for the opened factory
        /// </summary>
        protected override void OnOpening()
        {
            this.listenerTable = new FilterTable<UdpInputListener>();
            this.onReceive = new AsyncCallback(this.OnReceive);

            // create a buffer manager to re-use byte arrays as necessary for performance
            this.bufferManager = new BufferManager(UdpConstants.MaxBufferSize * 10, UdpConstants.MaxBufferSize);
        }

        /// <summary>
        /// Open the listener factory for use. Ensures our UDP socket is bound
        /// </summary>
        protected override void OnOpen()
        {
            base.OnOpen();
            if (this.listenSocket == null)
            {
                IPEndPoint endPoint = new IPEndPoint(IPAddress.Any, this.uri.Port);
                this.listenSocket = new Socket(endPoint.AddressFamily, SocketType.Dgram, ProtocolType.Udp);
                this.listenSocket.Bind(endPoint);
            }
        }

        /// <summary>
        /// Open has completed, start an asynchronous receive on our socket.
        /// </summary>
        protected override void OnOpened()
        {
            base.OnOpened();
            ThreadPool.QueueUserWorkItem(new WaitCallback(StartReceiving));
        }
        #endregion

        void StartReceiving(object state)
        {
            IAsyncResult result = null;

            try
            {
                lock (ThisLock)
                {
                    if (base.State == CommunicationState.Opened)
                    {
                        EndPoint dummy = new IPEndPoint(IPAddress.Any, 0);
                        byte[] buffer = this.bufferManager.TakeBuffer(UdpConstants.MaxBufferSize);
                        result = this.listenSocket.BeginReceiveFrom(buffer, 0, buffer.Length,
                            SocketFlags.None, ref dummy, this.onReceive, buffer);
                    }
                }

                if (result != null && result.CompletedSynchronously)
                {
                    ContinueReceiving(result);
                }
            }
            catch (Exception e)
            {
                Debug.WriteLine("Error in receiving from the socket.");
                Debug.WriteLine(e.ToString());
            }
        }

        void ContinueReceiving(IAsyncResult receiveResult)
        {
            bool continueReceiving = true;

            while (continueReceiving)
            {
                Message receivedMessage = null;

                if (receiveResult != null)
                {
                    receivedMessage = EndReceive(receiveResult);
                    receiveResult = null;
                }

                lock (ThisLock)
                {
                    if (base.State == CommunicationState.Opened)
                    {
                        EndPoint dummy = new IPEndPoint(IPAddress.Any, 0);
                        byte[] buffer = this.bufferManager.TakeBuffer(UdpConstants.MaxBufferSize);
                        receiveResult = this.listenSocket.BeginReceiveFrom(buffer, 0, buffer.Length,
                            SocketFlags.None, ref dummy, this.onReceive, buffer);
                    }
                }

                if (receiveResult == null || !receiveResult.CompletedSynchronously)
                {
                    continueReceiving = false;
                    Dispatch(receivedMessage);
                }
                else if (receivedMessage != null)
                {
                    ThreadPool.QueueUserWorkItem(new WaitCallback(DispatchCallback), receivedMessage);
                }
            }
        }

        Message EndReceive(IAsyncResult result)
        {
            // if we've started the shutdown process, then we've disposed
            // the socket and calls to socket.EndReceive will throw 
            if (base.State != CommunicationState.Opened)
                return null;

            byte[] buffer = (byte[])result.AsyncState;
            Debug.Assert(buffer != null);
            Message message = null;

            try
            {
                int count = 0;

                lock (ThisLock)
                {
                    // if we've started the shutdown process, socket is disposed
                    // and calls to socket.EndReceive will throw 
                    if (base.State == CommunicationState.Opened)
                    {
                        EndPoint dummy = new IPEndPoint(IPAddress.Any, 0);
                        count = this.listenSocket.EndReceiveFrom(result, ref dummy);
                    }
                }

                if (count > 0)
                {
                    message = UdpConstants.MessageEncoder.ReadMessage(new ArraySegment<byte>(buffer, 0, count), bufferManager);
                }
            }
            catch (Exception e)
            {
                Debug.WriteLine("Error in completing the async receive via EndReceiveFrom method.");
                Debug.WriteLine(e.ToString());
            }
            finally
            {
                if (message == null)
                {
                    this.bufferManager.ReturnBuffer(buffer);
                    buffer = null;
                }
            }

            return message;
        }

        //Called when an ansynchronous receieve operation completes
        //on the listening socket.
        void OnReceive(IAsyncResult result)
        {
            if (result.CompletedSynchronously)
                return;

            ContinueReceiving(result);
        }

        void DispatchCallback(object state)
        {
            Dispatch((Message)state);
        }

        /// <summary>
        /// Matches an incoming message to its waiting listener,
        /// using the FilterTable to dispatch the message to the correc
        /// listener. If no listener is waiting for the message, it is silently
        /// discarded.
        /// </summary>
        void Dispatch(Message message)
        {
            if (message == null)
                return;

            try
            {
                UdpInputListener listener = null;
                bool dispatchMessage = false;
                lock (ThisLock)
                {
                    dispatchMessage = this.listenerTable.Match(message, ref listener);
                }

                if (dispatchMessage)
                {
                    listener.Dispatch(message);
                }
                else
                {
                    try
                    {
                        EventLog.WriteEntry(UdpConstants.EventLogSourceName,
                            "A message was dropped because it didn't match any listeners.", EventLogEntryType.Warning);
                    }
                    catch
                    {
                    }
                }
            }
            catch (Exception e)
            {
                Debug.WriteLine("Error dispatching Message.");
                Debug.WriteLine(e.ToString());
            }
        }

        /// <summary>
        /// Used to get a unique uri (by CompositeDuplexChannelFactory for example).
        /// We get a unique TCP port by binding to "port 0"
        /// </summary>
        public override void SetUniqueUri(string host)
        {
            if (host == null)
            {
                throw new ArgumentNullException("host");
            }

            IPEndPoint endPoint = new IPEndPoint(IPAddress.Any, 0);
            int port;

            lock (this.ThisLock)
            {
                if (this.listenSocket != null)
                {
                    this.listenSocket.Close(0);
                }

                this.listenSocket = new Socket(endPoint.AddressFamily, SocketType.Dgram, ProtocolType.Udp);
                this.listenSocket.Bind(endPoint);
                port = ((IPEndPoint)this.listenSocket.LocalEndPoint).Port;
            }

            UriBuilder uriBuilder = new UriBuilder(Scheme, host, port);
            SetUri(uriBuilder.Uri);
        }

        public override void SetUniqueUri(string host, int port)
        {
            throw new NotSupportedException();
        }

        public override void SetUri(Uri baseAddress, string relativeAddress)
        {
            if (baseAddress == null)
                throw new ArgumentNullException("baseAddress");

            if (relativeAddress == null)
                throw new ArgumentNullException("relativeAddress");

            if (!baseAddress.IsAbsoluteUri)
                throw new ArgumentException(string.Format(CultureInfo.CurrentCulture, 
                    "Base address must be an absolute URI."), "baseAddress");

            if (baseAddress.Scheme != this.Scheme)
            {
                // URI schemes are case-insensitive, so try a case insensitive compare now
                if (string.Compare(baseAddress.Scheme, this.Scheme, true, System.Globalization.CultureInfo.InvariantCulture) != 0)
                {
                    throw new ArgumentException(string.Format(CultureInfo.CurrentCulture, 
                        "Invalid URI scheme: {0}.", baseAddress.Scheme), "baseAddress");
                }
            }

            Uri fullUri = baseAddress;

            // Ensure that baseAddress Path does end with a slash if we have a relative address
            if (relativeAddress != string.Empty)
            {
                if (!baseAddress.AbsolutePath.EndsWith("/"))
                {
                    UriBuilder uriBuilder = new UriBuilder(baseAddress);
                    uriBuilder.Path = uriBuilder.Path + "/";
                    baseAddress = uriBuilder.Uri;
                }

                fullUri = new Uri(baseAddress, relativeAddress);
            }

            lock (base.ThisLock)
            {
                ThrowIfDisposedOrImmutable();
                this.uri = fullUri;

                if (this.listenSocket != null)
                {
                    this.listenSocket.Close(0);
                    this.listenSocket = null;
                }
            }
        }

        /// <summary>
        /// Remove a listener from our filter table. Called when the listener is disposed.
        /// </summary>
        internal void UnregisterListener(UdpInputListener listener)
        {
            lock (this.ThisLock)
            {
                this.listenerTable.Remove(listener.Filter);
            }
        }
    }
}
