// ----------------------------------------------------------------------------
// Copyright (C) 2003-2005 Microsoft Corporation, All rights reserved.
// ----------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.ServiceModel;
using System.ServiceModel.Channels;
using System.Text;
using System.Xml;

namespace Microsoft.ServiceModel.Samples
{
    /// <summary>
    /// IOutputChannel implementation for Udp.
    /// </summary>
    class UdpOutputChannel : ChannelBase, IOutputChannel
    {
        IBufferManager bufferManager;
        EndpointAddress remoteAddress;
        EndPoint remoteEndPoint;
        Socket socket;

        internal UdpOutputChannel(UdpChannelFactory factory, EndpointAddress remoteAddress)
            : base(factory)
        {
            if (!string.Equals(remoteAddress.Uri.Scheme, factory.Scheme, StringComparison.InvariantCultureIgnoreCase))
            {
                throw new ArgumentException(string.Format(CultureInfo.CurrentCulture, 
                    "The scheme {0} specified in address is not supported.", remoteAddress.Uri.Scheme), "remoteAddress");
            }

            if (remoteAddress.Uri.IsDefaultPort)
            {
                throw new ArgumentException(string.Format(CultureInfo.CurrentCulture, 
                    "The address {0} must specify a remote port.", remoteAddress.Uri), "remoteAddress");
            }

            this.remoteAddress = remoteAddress;

            // convert the Uri host into an IP Address
            IPAddress remoteIP = null;

            switch (remoteAddress.Uri.HostNameType)
            {
                default:
                    throw new ArgumentException(string.Format(CultureInfo.CurrentCulture, 
                        "Can not determine the remote host address from {0}.",
                            this.remoteAddress.Uri.ToString()), "remoteAddress");

                case UriHostNameType.IPv4:
                case UriHostNameType.IPv6:
                        remoteIP = IPAddress.Parse(remoteAddress.Uri.Host);
                        break;

                case UriHostNameType.Basic:
                case UriHostNameType.Dns:
                    {
                        IPHostEntry hostEntry = Dns.GetHostEntry(remoteAddress.Uri.Host);
                        if (hostEntry.AddressList.Length > 0)
                        {
                            remoteIP = hostEntry.AddressList[0];
                        }
                        else
                        {
                            throw new ArgumentException(string.Format(CultureInfo.CurrentCulture, 
                                "Failed to resolve remote host: {0}.", remoteAddress.Uri.Host), 
                                "remoteAddress");
                        }
                        break;
                    }
            }

            this.remoteEndPoint = new IPEndPoint(remoteIP, remoteAddress.Uri.Port);
            this.bufferManager = factory.BufferManager;
        }

        public EndpointAddress RemoteAddress
        {
            get
            {
                return this.remoteAddress;
            }
        }

#pragma warning disable 67
        public event EventHandler RemoteAddressChanged;
#pragma warning restore 67

        public IAsyncResult BeginSend(Message message, AsyncCallback callback, object state)
        {
            base.ThrowIfDisposedOrNotOpen();
            return new SendAsyncResult(this, message, callback, state);
        }

        /// <summary>
        /// Address the Message and serialize it into a byte array. We use a hardcoded MessageEncoderFactory 
        /// for Text(SOAP 1.2).  Alternative implementations could ahve this be configurable, but then that 
        /// information would beed to be coordinated between the two endpoints involved (either in a framing
        /// protocol or out of band).
        /// </summary>
        ArraySegment<byte> EncodeMessage(Message message)
        {
            try
            {
                this.remoteAddress.ApplyTo(message);
                return UdpConstants.MessageEncoder.WriteMessage(message, UdpConstants.MaxBufferSize, bufferManager);
            }
            finally
            {
                // we've consumed the message by serializing it, so clean up
                message.Close();
            }
        }

        public void EndSend(IAsyncResult result)
        {
            SendAsyncResult.End(result);
        }

        /// <summary>
        /// Shutdown ungracefully
        /// </summary>
        protected override void OnAbort()
        {
            try
            {
                this.socket.Close(0);
            }
            finally
            {
                base.OnAbort();
            }
        }

        /// <summary>
        /// Shutdown gracefully
        /// </summary>
        protected override void OnClose()
        {
            this.socket.Close();
            base.OnClose();
        }

        /// <summary>
        /// Open the channel for use. This creates a new UDP socket to use for sends.
        /// </summary>
        protected override void OnOpen()
        {
            this.socket = new Socket(this.remoteEndPoint.AddressFamily, SocketType.Dgram, ProtocolType.Udp);
        }

        public void Send(Message message)
        {
            base.ThrowIfDisposedOrNotOpen();

            ArraySegment<byte> messageBuffer = EncodeMessage(message);

            try
            {
                int bytesSent = this.socket.SendTo(messageBuffer.Array, messageBuffer.Offset, messageBuffer.Count,
                    SocketFlags.None, this.remoteEndPoint);

                if (bytesSent != messageBuffer.Count)
                {
                    throw new ChannelIOException(string.Format(CultureInfo.CurrentCulture, 
                        "A Udp error occurred sending a message to {0}.", this.remoteEndPoint));
                }
            }
            catch (SocketException socketException)
            {
                throw UdpChannelHelpers.ConvertTransferException(socketException);
            }
            finally
            {
                // we need to make sure buffers are always returned to the IBufferManager
                bufferManager.ReturnBuffer(messageBuffer.Array);
            }
        }

        /// <summary>
        /// Implementation of async send for Udp. 
        /// </summary>
        class SendAsyncResult : AsyncResult
        {
            ArraySegment<byte> messageBuffer;
            UdpOutputChannel channel;

            public SendAsyncResult(UdpOutputChannel channel, Message message, AsyncCallback callback, object state)
                : base(callback, state)
            {
                this.channel = channel;
                this.messageBuffer = channel.EncodeMessage(message);
                try
                {
                    IAsyncResult result = null;
                    try
                    {
                        result = channel.socket.BeginSendTo(messageBuffer.Array, messageBuffer.Offset, messageBuffer.Count,
                            SocketFlags.None, channel.remoteEndPoint, new AsyncCallback(OnSend), this);
                    }
                    catch (SocketException socketException)
                    {
                        throw UdpChannelHelpers.ConvertTransferException(socketException);
                    }

                    if (!result.CompletedSynchronously)
                        return;

                    CompleteSend(result, true);
                }
                catch
                {
                    CleanupBuffer();
                    throw;
                }
            }

            void CleanupBuffer()
            {
                if (messageBuffer.Array != null)
                {
                    this.channel.bufferManager.ReturnBuffer(messageBuffer.Array);
                    messageBuffer = new ArraySegment<byte>();
                }
            }

            void CompleteSend(IAsyncResult result, bool synchronous)
            {
                try
                {
                    int bytesSent = channel.socket.EndSendTo(result);

                    if (bytesSent != messageBuffer.Count)
                    {
                        throw new ChannelIOException(string.Format(CultureInfo.CurrentCulture, 
                            "A Udp error occurred sending a message to {0}.", channel.remoteEndPoint));
                    }
                }
                catch (SocketException socketException)
                {
                    throw UdpChannelHelpers.ConvertTransferException(socketException);
                }
                finally
                {
                    CleanupBuffer();
                }

                base.Complete(synchronous);
            }

            void OnSend(IAsyncResult result)
            {
                if (result.CompletedSynchronously)
                    return;

                try
                {
                    CompleteSend(result, false);
                }
                catch (Exception e)
                {
                    base.Complete(false, e);
                }
            }

            public static void End(IAsyncResult result)
            {
                SendAsyncResult sendResult = result as SendAsyncResult;

                if (result == null)
                    throw new ArgumentException("Invalid AsyncResult.");

                AsyncResult.End(sendResult);
            }
        }
    }
}
